///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hdate_time;

import 'package:haystack/src/string_utils.dart';
import 'package:quiver/core.dart';

import 'HDate.dart';
import 'HTime.dart';
import 'HTimeZone.dart';
import 'HVal.dart';
import 'ParseError.dart';

/// HDateTime models a timestamp with a specific timezone.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds).
class HDateTime extends HVal {
  factory HDateTime.fromValues(
      int year, int month, int day, int hour, int min, int sec, HTimeZone tz)
  {
    HDate date = HDate(year, month, day);
    HTime time = HTime(hour, min, sec);

    return HDateTime(date, time, tz);
  }

  factory HDateTime.fromMillis(int ms, HTimeZone tz) {
    return HDateTime.fromDateTime(
        DateTime.fromMillisecondsSinceEpoch(ms, isUtc: true), tz);
  }

  factory HDateTime.fromDateTime(DateTime dt, HTimeZone tz) {
    // if (tz == null) {
    //   tz = HTimeZone.DEFAULT;
    // }
    DateTime fixedTime = dt.add(Duration(seconds: tz.offset));
    return HDateTime(HDate.fromDateTime(fixedTime),
        HTime.fromDateTime(fixedTime), tz);
  }

  factory HDateTime.fromString(String s) {
    // date
    int tIdx = s.indexOf('T');
    if (tIdx < 0) throw ParseError(s);
    HDate date = HDate.fromString(s.substring(0, tIdx));

    // find timezone offset and name indices.
    // this, in turn, allows us to find the time
    if (s.endsWith('Z')) {
      return HDateTime(
          date,
          HTime.fromString(s.substring(tIdx + 1, s.length - 1)),
          HTimeZone.UTC);
    } else if (s.endsWith('Z UTC')) {
      return HDateTime(
          date,
          HTime.fromString(
              s.substring(tIdx + 1, s.length - 'Z UTC'.length)),
          HTimeZone.UTC);
    }

    // have tz offset and name
    int spIdx = s.indexOf(' ');
    if (spIdx < 0) throw ParseError('Expected time zone name: $s');
    // work backward from space to find beginning of offset
    int offsetIdx = spIdx - 1;
    while (true) {
      if (offsetIdx <= tIdx) {
        throw ParseError('Expected Z or -/+ for timezone offset: $s');
      }
      int c = s.codeUnitAt(offsetIdx);
      if (c == MINUS_CHAR || c == PLUS_CHAR || c == Z_CHAR) break;
      --offsetIdx;
    }
    String offsetStr = s.substring(offsetIdx, spIdx);
    int offset = 'Z' == offsetStr ? 0 : _parseOffset(offsetStr);
    HTimeZone tz = HTimeZone(s.substring(spIdx + 1), offset);

    return HDateTime(
        date, HTime.fromString(s.substring(tIdx + 1, offsetIdx)), tz);
  }

  factory HDateTime.now(HTimeZone tz) =>
      HDateTime.fromDateTime(DateTime.now(), tz);

  HDateTime(this.date, this.time, this.tz);


  /// Date component of the timestamp
  final HDate date;

  /// Time component of the timestamp
  final HTime time;

  /// Offset in seconds from UTC including DST offset
  int get tzOffset => tz.offset;

  /// Timezone as Olson database city name
  final HTimeZone tz;

  /// Get this date time as Java milliseconds since epoch
  int millis() {
    return DateTime.utc(date.year, date.month, date.day, time.hour,
            time.min, time.sec, time.ms)
        .subtract(Duration(seconds: tz.offset))
        .millisecondsSinceEpoch;
  }

  /// Hash is based on date, time, tzOffset, and (NOT) tz
  @override
  int get hashCode => _hashCode ?? (hash3(date.hashCode, time.hashCode, tz.hashCode));
  int? _hashCode;

  @override
  bool operator ==(Object that) {
    return that is HDateTime &&
        date == that.date &&
        time == that.time &&
        tz == that.tz &&
        tz.name == that.tz.name;
  }

  /// Comparison based on millis.
  @override
  int compareTo(Object that) {
    int thisMillis = this.millis();
    int thatMillis = (that as HDateTime).millis();
    if (thisMillis < thatMillis) {
      return -1;
    } else if (thisMillis > thatMillis) {
      return 1;
    }
    return 0;
  }

  /// Encodes as 't:YYYY-MM-DD'T'hh:mm:ss.FFFz zzzz'.
  @override
  String toJson() {
    StringBuffer s = StringBuffer();
    s.write('t:');
    _encode(s);
    return s.toString();
  }

  /// Encodes as 'YYYY-MM-DD'T'hh:mm:ss.FFFz zzzz'.
  @override
  String toZinc() {
    StringBuffer s = StringBuffer();
    _encode(s);
    return s.toString();
  }

  static int _parseOffset(String s) {
    if (s.length != '-HH:MM'.length) {
      throw ParseError('Invalid tz offset: $s');
    }
    int sign = s.startsWith('-') ? -1 : 1;
    int tzHours = int.parse(s.substring(1, 3));
    if (s.codeUnitAt(3) != COLON_CHAR) {
      throw ParseError('Invalid tz offset: $s');
    }
    int tzMins = int.parse(s.substring(4));
    return sign * (tzHours * 3600) + (tzMins * 60);
  }

  void _encode(StringBuffer s) {
    date.encode(s);
    s.write('T');
    time.encode(s);
    if (tz.offset == 0) {
      s.write('Z');
    } else {
      int offset = this.tz.offset; //seconds
      if (offset < 0) {
        s.write('-');
        offset = -offset;
      } else {
        s.write('+');
      }

      int zh = offset ~/ 3600;
      int zm = (offset % 3600) ~/ 60;
      if (zh < 10) {
        s.write('0');
      }
      s..write(zh)..write(':');
      if (zm < 10) {
        s.write('0');
      }
      s.write(zm);
    }

    s..write(' ')..write(tz);
  }
}
