///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.val_test;

import 'package:haystack/haystack.dart';
import 'package:test/test.dart';

import 'utils.dart';

void main() {
  group('Test marker', () {
    test('equality', () {
      expect(HMarker.VAL, equals(HMarker.VAL));
    });

    test('toString', () {
      expect(HMarker.VAL.toString(), equals('marker'));
    });

    test('toZinc', () {
      expect(HMarker.VAL.toZinc(), equals('M'));
    });
  });

  group('Test HBool', () {
    test('equality', () {
      expect(HBool.TRUE == HBool.TRUE, true);
      expect(HBool.FALSE == HBool.FALSE, true);
      expect(HBool.TRUE == HBool.FALSE, false);
      expect(identical(HBool(true), HBool.TRUE), true);
      expect(identical(HBool(false), HBool.FALSE), true);
    });
    test('compare', () {
      expect(HBool.FALSE.compareTo(HBool.TRUE) < 0, true);
      expect(HBool.TRUE.compareTo(HBool.TRUE) == 0, true);
    });
    test('toString', () {
      expect(HBool.TRUE.toString(), equals('true'));
      expect(HBool.FALSE.toString(), equals('false'));
    });
    test('toZinc', () {
      verifyZinc(HBool.TRUE, 'T');
      verifyZinc(HBool.FALSE, 'F');
    });
  });

  group('Test HNum', () {
    test('equality', () {
      expect(HNum(2) == HNum(2.0, null), true);
      expect(HNum(2) == HNum(2, '%'), false);
      expect(HNum(2, '%') == HNum(2), false);
      expect(HNum(2, '%') == HNum(2, '%'), true);
      expect(identical(HNum(0), HNum(0.0, null)), true);
      expect(HNum(double.parse('-NaN')) == HNum(double.nan, null), true);
    });

    test('compare', () {
      expect(HNum(9).compareTo(HNum(11)) < 0, true);
      expect(HNum(-3).compareTo(HNum(-4)) > 0, true);
      expect(HNum(-23).compareTo(HNum(-23)) == 0, true);
    });

    test('zinc', () {
      verifyZinc(HNum(123), '123');
      verifyZinc(HNum(123.4, 'm/s'), '123.4m/s');
      verifyZinc(HNum(9.6, 'm/s'), '9.6m/s');
      verifyZinc(HNum(-5.2, '\u00b0F'), '-5.2\u00b0F');
      verifyZinc(HNum(23, '%'), '23%');
      verifyZinc(HNum(2.4e-3, 'fl_oz'), "0.0024fl_oz");
      verifyZinc(HNum(2.4e5, "\$"), "240000\$");
      expect(read("1234.56fl_oz"), HNum(1234.56, "fl_oz"));
      expect(read("0.000028fl_oz"), HNum(0.000028, "fl_oz"));
    });

    test('zinc specials', () {
      expect(HNum(double.nan).toZinc(), equals('NaN'));
      expect(HNum(double.negativeInfinity).toZinc(), equals('-INF'));
      expect(HNum(double.infinity).toZinc(), equals('INF'));
    });

    test('zinc verify units never serialized for special values', () {
      expect(HNum(double.nan, 'ignore').toZinc(), equals('NaN'));
      expect(HNum(double.negativeInfinity, '%').toZinc(), equals('-INF'));
      expect(HNum(double.infinity, '%').toZinc(), equals('INF'));
    });

    test('verify bad unit names', () {
      expect(HNum.isUnitName(null), true);
      expect(HNum.isUnitName(''), false);
      expect(HNum.isUnitName('x_z'), true);
      expect(HNum.isUnitName('x z'), false);
      expect(() => HNum(123.4, 'foo bar'), throwsA(const TypeMatcher<ArgumentError>()));
      expect(() => HNum(123.4, 'foo,bar'), throwsA(const TypeMatcher<ArgumentError>()));
    });

    //todo check locale change test
  });

  group('Test HStr', () {
    test('equality', () {
      expect(HUri('a') == HUri('a'), true);
      expect(HUri('a') == HUri('b'), false);
      expect(identical(HUri(''), HUri('')), true);
    });

    test('compare', () {
      expect(HUri('abc').compareTo(HUri('z')) < 0, true);
      expect(HUri('Foo').compareTo(HUri('Foo')) == 0, true);
    });

    test('zinc', () {
      verifyZinc(HStr("hello"), "\"hello\"");
      verifyZinc(HStr("_ \\ \" \n \r \t \u0011 _"), "\"_ \\\\ \\\" \\n \\r \\t \\u0011 _\"");
      verifyZinc(HStr("\u0abc"), "\"\u0abc\"");
    });

    test('hex upper and lower case', () {
      expect(read("\"[\\uabcd \\u1234]\""), equals(HStr("[\uabcd \u1234]")));
      expect(read("\"[\\uABCD \\u1234]\""), equals(HStr("[\uABCD \u1234]")));
      expect(() => read("\"end..."), throwsA(anything));
      expect(() => read("\"\\u1x34\""), throwsA(const TypeMatcher<ParseError>()));
    });
  });

  group('Test HUri', () {
    test('equality', () {
      expect(HUri('a') == HUri('a'), true);
      expect(HUri('a') == HUri('b'), false);
      expect(identical(HUri(''), HUri('')), true);
    });

    test('compare', () {
      expect(HUri('abc').compareTo(HUri('z')) < 0, true);
      expect(HUri('Foo').compareTo(HUri('Foo')) == 0, true);
    });

    test('zinc', () {
      verifyZinc(HUri("http://foo.com/f?q"), "`http://foo.com/f?q`");
      verifyZinc(HUri("a\$b"), "`a\$b`");
      verifyZinc(HUri("a`b"), "`a\\`b`");
      verifyZinc(HUri("http\\:a\\?b"), "`http\\:a\\?b`");
      verifyZinc(HUri("\u01ab.txt"), "`\u01ab.txt`");
    });

    test('errors', () {
      expect(() => read("`no end"), throwsA(const TypeMatcher<ParseError>()));
      expect(
          () => read("`new\nline`"), throwsA(const TypeMatcher<ParseError>()));
    });
  });

  group('Test HRef', () {
    test('equality (ignore [dis])', () {
      expect(HRef('foo') == HRef('foo'), true);
      expect(HRef('foo') == HRef('foo', 'Foo'), true);
      expect(HRef('foo') == HRef('Foo'), false);
    });

    test('zinc', () {
      verifyZinc(HRef("1234-5678.foo:bar"), "@1234-5678.foo:bar");
      verifyZinc(HRef("1234-5678", "Foo Bar"), "@1234-5678 \"Foo Bar\"");
      verifyZinc(HRef("1234-5678", "Foo \"Bar\""),
          "@1234-5678 \"Foo \\\"Bar\\\"\"");
    });

    test('verify bad refs are caught on encoding', () {
      expect(HRef.isId(""), false);
      expect(HRef.isId("%"), false);
      expect(HRef.isId("a"), true);
      expect(HRef.isId("a-b:c"), true);
      expect(HRef.isId("a b"), false);
      expect(HRef.isId("a\u0129b"), false);
      expect(() => HRef("@a"), throwsA(anything));
      expect(() => HRef("a b"), throwsA(anything));
      expect(() => HRef("a\n"), throwsA(anything));
      expect(() => read("@"), throwsA(anything));
    });
  });

  group('Test HBin', () {
    test('equality', () {
      expect(HBin('text/plain') == HBin('text/plain'), true);
      expect(HBin('text/plain') == HBin('text/xml'), false);
    });

    test('toZinc', () {
      verifyZinc(HBin("text/plain"), 'Bin("text/plain")');
      verifyZinc(HBin("text/plain; charset=utf-8"),
          'Bin("text/plain; charset=utf-8")');
    });

    test('verify bad bins are caught on encoding', () {
      expect(() => HBin("text/plain; f()").toZinc(), throwsA(anything));
      expect(() => read("Bin()"), throwsA(anything));
      expect(() => read("Bin(text)"), throwsA(anything));
    });
  });

  group('Test HDate', () {
    test('equality', () {
      expect(HDate(2011, 6, 7) == HDate(2011, 6, 7), true);
      expect(HDate(2011, 6, 7) == HDate(2011, 6, 8), false);
      expect(HDate(2011, 6, 7) == HDate(2011, 2, 7), false);
      expect(HDate(2011, 6, 7) == HDate(2009, 6, 7), false);
    });

    test('compare', () {
      expect(HDate(2011,  6, 9).compareTo(HDate(2011, 6, 21))  < 0, true);
      expect(HDate(2011, 10, 9).compareTo(HDate(2011, 3, 21))  > 0, true);
      expect(HDate(2010,  6, 9).compareTo(HDate(2000, 9, 30))  > 0, true);
      expect(HDate(2010,  6, 9).compareTo(HDate(2010, 6,  9)) == 0, true);
    });

    test('plus/minus', () {
      expect(HDate(2011, 12, 1).minusDays(0)  == HDate(2011, 12,  1), true);
      expect(HDate(2011, 12, 1).minusDays(1)  == HDate(2011, 11, 30), true);
      expect(HDate(2011, 12, 1).minusDays(-2) == HDate(2011, 12,  3), true);
      expect(HDate(2011, 12, 1).plusDays(2)   == HDate(2011, 12,  3), true);
      expect(HDate(2011, 12, 1).plusDays(31)  == HDate(2012, 1,   1), true);
      expect(HDate(2008,  3, 3).minusDays(3)  == HDate(2008, 2,  29), true);
      expect(HDate(2008,  3, 3).minusDays(4)  == HDate(2008, 2,  28), true);
    });

    test('zinc', () {
      verifyZinc(HDate(2011, 6, 7), '2011-06-07');
      verifyZinc(HDate(2011, 10, 10), '2011-10-10');
      verifyZinc(HDate(2011, 12, 31), "2011-12-31");
      expect(() => read("2003-xx-02"), throwsA(anything));
      expect(() => read("2003-02"), throwsA(anything));
      expect(() => read("2003-02-xx"), throwsA(anything));
    });

    test('leap year', () {
      for (int y = 1900; y <= 2100; y++) {
        if (((y % 4) == 0) && (y != 1900) && (y != 2100)) {
          expect(HDate.isLeapYear(y), true);
        } else {
          expect(HDate.isLeapYear(y), false);
        }
      }
    });
  });

  group('Test HTime', () {
    test('equality', () {
      expect(HTime(1, 2, 3, 4) == HTime(1, 2, 3, 4), true);
      expect(HTime(1, 2, 3, 4) == HTime(9, 2, 3, 4), false);
      expect(HTime(1, 2, 3, 4) == HTime(1, 9, 3, 4), false);
      expect(HTime(1, 2, 3, 4) == HTime(1, 2, 9, 9), false);
    });

    test('compare', () {
      expect(HTime(0, 0, 0, 0).compareTo(HTime(0, 0, 0, 9)) < 0, true);
      expect(HTime(0, 0, 0, 0).compareTo(HTime(0, 0, 1, 0)) < 0, true);
      expect(HTime(0, 1, 0, 0).compareTo(HTime(0, 0, 0, 0)) > 0, true);
      expect(HTime(0, 0, 0, 0).compareTo(HTime(2, 0, 0, 0)) < 0, true);
      expect(HTime(2, 0, 0, 0).compareTo(HTime(2, 0, 0, 0)) == 0, true);
    });

    test('zinc', () {
      verifyZinc(HTime(2, 3), '02:03:00');
      verifyZinc(HTime(2, 3, 4), '02:03:04');
      verifyZinc(HTime(2, 3, 4, 5), "02:03:04.005");
      verifyZinc(HTime(2, 3, 4, 56), "02:03:04.056");
      verifyZinc(HTime(2, 3, 4, 109), "02:03:04.109");
      verifyZinc(HTime(2, 3, 10, 109), "02:03:10.109");
      verifyZinc(HTime(2, 10, 59), "02:10:59");
      verifyZinc(HTime(10, 59, 30), "10:59:30");
      verifyZinc(HTime(23, 59, 59, 999), "23:59:59.999");
    });

    test('errors', () {
      //expect(() => read("3:20:00"), throwsA(anything));
      expect(() => read("13:xx:00"), throwsA(anything));
      expect(() => read("13:45:0x"), throwsA(anything));
      // expect(() => read("13:45:00.4561"), throwsA(anything)); fixme should throws
    });
  });

  group('Test HDateTime', () {
    HTimeZone utc = HTimeZone.UTC;
    HTimeZone london = HTimeZone("London", 0);
    HTimeZone london2 = HTimeZone("London", 3600);

    test('equality', () {
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc), true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2009, 1, 2, 3, 4, 5, utc), false);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 9, 2, 3, 4, 5, utc), false);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 1, 9, 3, 4, 5, utc), false);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 1, 2, 9, 4, 5, utc), false);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 1, 2, 3, 9, 5, utc), false);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 1, 2, 3, 4, 9, utc), false);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 1, 2, 3, 4, 5, london), false);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc) == HDateTime.fromValues(2011, 1, 2, 3, 4, 5, london2), false);
    });

    test('compare', () {
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc)) == 0, true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2011, 1, 2, 3, 4, 6, utc)) < 0, true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2011, 1, 2, 3, 5, 5, utc)) < 0, true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2011, 1, 2, 4, 4, 5, utc)) < 0, true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2011, 1, 3, 3, 4, 5, utc)) < 0, true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2011, 2, 2, 3, 4, 5, utc)) < 0, true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2012, 1, 2, 3, 4, 5, utc)) < 0, true);
      expect(HDateTime.fromValues(2011, 1, 2, 3, 4, 5, utc).compareTo(HDateTime.fromValues(2011, 1, 2, 3, 4, 0, utc)) > 0, true);
    });

    test("zinc", () {
      HDateTime ts = HDateTime.fromMillis(1307377618069, HTimeZone("New_York", tzOffset(-4, 0)));
      verifyZinc(ts, "2011-06-06T12:26:58.069-04:00 New_York");
      expect(ts.date.toString(), equals("2011-06-06"));
      expect(ts.time.toString(), equals("12:26:58.069"));
      expect(ts.tz.offset, equals(-4 * 60 * 60));
      expect(ts.tz.name, equals("New_York"));
//      expect(ts.tz.java.getID(), "America/New_York"); // Dart hasn't this funcs
      expect(ts.millis(), equals(1307377618069));

      // convert back to millis
      ts = HDateTime(ts.date, ts.time, ts.tz);
      expect(ts.millis(), equals(1307377618069));

      // different timezones
      ts = HDateTime.fromMillis(949478640000, HTimeZone("New_York", -18000));
      verifyZinc(ts, "2000-02-02T03:04:00-05:00 New_York");
      ts = HDateTime.fromMillis(949478640000, HTimeZone("UTC", 0));
      verifyZinc(ts, "2000-02-02T08:04:00Z UTC");
      ts = HDateTime.fromMillis(949478640000, HTimeZone("Taipei", 28800));
      verifyZinc(ts, "2000-02-02T16:04:00+08:00 Taipei");
      verifyZinc(HDateTime.fromValues(2011, 6, 7, 11, 3, 43, HTimeZone("GMT+10", -36000)),
          "2011-06-07T11:03:43-10:00 GMT+10");
      verifyZinc(HDateTime(HDate(2011, 6, 8), HTime(4, 7, 33, 771), HTimeZone("GMT-7", 25200)),
          "2011-06-08T04:07:33.771+07:00 GMT-7");
    });

    test('verify millis()', () {
      HDate date = HDate(2014, 12, 24);
      HTime time = HTime(11, 12, 13, 456);
      HTimeZone newYork = HTimeZone("New_York", -18000);
      int utcMillis = 1419437533456;

      HDateTime a = HDateTime(date, time, newYork);
      HDateTime b = HDateTime.fromMillis(utcMillis, newYork);
      HDateTime c = HDateTime.fromString("2014-12-24T11:12:13.456-05:00 New_York");

      expect(a.millis(), equals(utcMillis));
      expect(b.millis(), equals(utcMillis));
      expect(c.millis(), equals(utcMillis));
    });

    test('errors', () {
      expect(() => read("2000-02-02T03:04:00-0x:00 New_York"), throwsA(anything));
      expect(() => read("2000-02-02T03:04:00-05 New_York"), throwsA(anything));
      expect(() => read("2000-02-02T03:04:00-05:!0 New_York"), throwsA(anything));
      expect(() => read("2000-02-02T03:04:00-05:00"), throwsA(anything));
      expect(() => read("2000-02-02T03:04:00-05:00 @"), throwsA(anything));
    });
  });

  test('Test Midnight', (){
    verifyMidnight(HDate(2011, 11, 3),  "UTC",      0,      "2011-11-03T00:00:00Z UTC");
    verifyMidnight(HDate(2011, 11, 3),  "New_York", -14400, "2011-11-03T00:00:00-04:00 New_York");
    verifyMidnight(HDate(2011, 12, 15), "Chicago",  -21600, "2011-12-15T00:00:00-06:00 Chicago");
    verifyMidnight(HDate(2008, 2, 29),  "Phoenix", -25200,  "2008-02-29T00:00:00-07:00 Phoenix");
  });

  group('Test HCoord', () {
    test('first', () {
      verifyCoord(12, 34, "C(12.0,34.0)");
    });

    test('lat boundaries', () {
      verifyCoord(90, 123, "C(90.0,123.0)");
      verifyCoord(-90, 123, "C(-90.0,123.0)");
      verifyCoord(89.888999, 123, "C(89.888999,123.0)");
      verifyCoord(-89.888999, 123, "C(-89.888999,123.0)");
    });

    test('lon boundaries', () {
      verifyCoord(45, 180, "C(45.0,180.0)");
      verifyCoord(45, -180, "C(45.0,-180.0)");
      verifyCoord(45, 179.999129, "C(45.0,179.999129)");
      verifyCoord(45, -179.999129, "C(45.0,-179.999129)");
    });

    test('decimal places', () {
      verifyCoord(9.1, -8.1, "C(9.1,-8.1)");
      verifyCoord(9.12, -8.13, "C(9.12,-8.13)");
      verifyCoord(9.123, -8.134, "C(9.123,-8.134)");
      verifyCoord(9.1234, -8.1346, "C(9.1234,-8.1346)");
      verifyCoord(9.12345, -8.13456, "C(9.12345,-8.13456)");
      verifyCoord(9.123452, -8.134567, "C(9.123452,-8.134567)");
    });

    test('zero boundaries', () {
      verifyCoord(0, 0, "C(0.0,0.0)");
      verifyCoord(0.3, -0.3, "C(0.3,-0.3)");
      verifyCoord(0.03, -0.03, "C(0.03,-0.03)");
      verifyCoord(0.003, -0.003, "C(0.003,-0.003)");
      verifyCoord(0.0003, -0.0003, "C(0.0003,-0.0003)");
      verifyCoord(0.02003, -0.02003, "C(0.02003,-0.02003)");
      verifyCoord(0.020003, -0.020003, "C(0.020003,-0.020003)");
      verifyCoord(0.000123, -0.000123, "C(0.000123,-0.000123)");
      verifyCoord(7.000123, -7.000123, "C(7.000123,-7.000123)");
    });

    test('arg errors', () {
      expect(HCoord.isLat(-91.0), false);
      expect(HCoord.isLat(-90.0), true);
      expect(HCoord.isLat(-89.0), true);
      expect(HCoord.isLat(90.0), true);
      expect(HCoord.isLat(91.0), false);
      expect(HCoord.isLng(-181.0), false);
      expect(HCoord.isLng(-179.99), true);
      expect(HCoord.isLng(180.0), true);
      expect(HCoord.isLng(181.0), false);
      expect(() => HCoord(91, 12), throwsA(const TypeMatcher<ArgumentError>()));
      expect(() => HCoord(-90.2, 12), throwsA(const TypeMatcher<ArgumentError>()));
      expect(() => HCoord(13, 180.009), throwsA(const TypeMatcher<ArgumentError>()));
      expect(() => HCoord(91, -181), throwsA(const TypeMatcher<ArgumentError>()));
    });

    test('parse errs', () {
      expect(() => HCoord.fromString("1.0,2.0"), throwsA(const TypeMatcher<ParseError>()));
      expect(() => HCoord.fromString("(1.0,2.0)"), throwsA(const TypeMatcher<ParseError>()));
      expect(() => HCoord.fromString("C(1.0,2.0"), throwsA(const TypeMatcher<ParseError>()));
      expect(() => HCoord.fromString("C(x,9)"), throwsA(const TypeMatcher<ParseError>()));
    });
  });

  group('Test HRange', () {
    HTimeZone ny = HTimeZone("New_York", -14400);
    HDate today = HDate.today(ny);

    //@Skip("http://project-haystack.org/forum/topic/450")
    test('no name', () {
      HDate yesterday = today.minusDays(1);
      HDate x = HDate(2011, 7, 4);
      HDate y = HDate(2011, 11, 4);
      HDateTime xa = HDateTime(x, HTime(2, 30), ny);
      HDateTime xb = HDateTime(x, HTime(22, 5), ny);

      verifyRange(HDateTimeRange.fromString("today", ny), today, today);
      verifyRange(
          HDateTimeRange.fromString("yesterday", ny), yesterday, yesterday);
      verifyRange(HDateTimeRange.fromString("2011-07-04", ny), x, x);
      verifyRange(
          HDateTimeRange.fromString("2011-07-04,2011-11-04", ny), x, y);
      verifyRange2(HDateTimeRange.fromString("$xa,$xb", ny), xa, xb);

      HDateTimeRange r = HDateTimeRange.fromString(xb.toString(), ny);
      expect(r.start, xb);
      expect(r.end.date, today);
      expect(r.end.tz, ny);
    });

    test('this week', () {
      HDate mon = today;
      HDate sun = today;
      while (mon.weekday() > DateTime.monday) {
        mon = mon.minusDays(1);
      }
      while (sun.weekday() < DateTime.sunday) {
        sun = sun.plusDays(1);
      }
      verifyRange(HDateTimeRange.thisWeek(ny), mon, sun);
    });

    test('this month', () {
      HDate first = today;
      HDate last = today;
      while (first.day > 1) {
        first = first.minusDays(1);
      }
      while (last.day < HDate.daysInMonth(today.year, today.month)) {
        last = last.plusDays(1);
      }
      verifyRange(HDateTimeRange.thisMonth(ny), first, last);
    });

    test('this year', () {
      HDate first = HDate(today.year, 1, 1);
      HDate last = HDate(today.year, 12, 31);
      verifyRange(HDateTimeRange.thisYear(ny), first, last);
    });

    test('last week', () {
      HDate prev = today.minusDays(7);
      HDate mon = prev;
      HDate sun = prev;
      while (mon.weekday() > DateTime.monday) {
        mon = mon.minusDays(1);
      }
      while (sun.weekday() < DateTime.sunday) {
        sun = sun.plusDays(1);
      }
      verifyRange(HDateTimeRange.lastWeek(ny), mon, sun);
    });

    test('last month', () {
      HDate last = today;
      while (last.month == today.month) {
        last = last.minusDays(1);
      }
      HDate first = HDate(last.year, last.month, 1);
      verifyRange(HDateTimeRange.lastMonth(ny), first, last);
    });

    test('last year', () {
      HDate first = HDate(today.year - 1, 1, 1);
      HDate last = HDate(today.year - 1, 12, 31);
      verifyRange(HDateTimeRange.lastYear(ny), first, last);
    });
  });
}

/**
 * Testing creation, parameters, equality and string parsing
 */
void verifyCoord(num lat, num lng, String s) {
  HCoord c = HCoord(lat, lng);
  expect(c.lat(), equals(lat));
  expect(c.lng(), equals(lng));
  expect(c.toString(), equals(s));
  expect(HCoord.fromString(s), equals(c));
}

void verifyMidnight(HDate date, String tzName, int tzOffset, String str) {
  HDateTime ts = date.midnight(HTimeZone(tzName, tzOffset));
  expect(ts.date, equals(date));
  expect(ts.time.hour, equals(0));
  expect(ts.time.min, equals(0));
  expect(ts.time.sec, equals(0));
  expect(ts.toString(), equals(str));
  expect(ts, equals(read(ts.toZinc())));
  expect(ts.millis(), equals((read(str) as HDateTime).millis()));
}

void verifyRange(HDateTimeRange r, HDate start, HDate end) {
  expect(r.start.date, equals(start));
  expect(r.start.time, equals(HTime.MIDNIGHT));
  expect(r.start.tz.name, equals("New_York"));
  expect(r.end.date, equals(end.plusDays(1)));
  expect(r.end.time, equals(HTime.MIDNIGHT));
  expect(r.end.tz.name, equals("New_York"));
}

void verifyRange2(HDateTimeRange r, HDateTime start, HDateTime end) {
  expect(r.start, equals(start));
  expect(r.end, equals(end));
}
